"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const builder_1 = require("./builder");
const accessIdentity_1 = require("./distribution/accessIdentity");
const gwTarget_1 = require("./distribution/gwTarget");
const index_1 = require("./distribution/index");
const s3Target_1 = require("./distribution/s3Target");
const bucket_1 = require("./s3/bucket");
const permission_1 = require("./s3/permission");
function generate(app) {
    const sls = new builder_1.default(app.name);
    for (const api of app.backend) {
        const { method, path } = api.module;
        let events;
        if (path.endsWith('*')) {
            const proxyPath = path.replace(/\/+\*$/, '');
            events = [
                { http: { method, path: `${proxyPath}/` } },
                { http: { method, path: `${proxyPath}/{proxy+}` } },
            ];
        }
        else {
            events = [{ http: { method, path } }];
        }
        const fn = { handler: api.handler, timeout: 30, events };
        sls.addFunction(api.name, fn);
    }
    for (const model of app.models) {
        const table = {
            Properties: {
                AttributeDefinitions: [
                    {
                        AttributeName: 'id',
                        AttributeType: 'S',
                    },
                ],
                KeySchema: [
                    {
                        AttributeName: 'id',
                        KeyType: 'HASH',
                    },
                ],
                ProvisionedThroughput: {
                    ReadCapacityUnits: model.reads,
                    WriteCapacityUnits: model.writes,
                },
                TableName: `${model.appName}-${model.name}`,
            },
            Type: 'AWS::DynamoDB::Table',
        };
        sls.addTable(model.name, table);
    }
    const distributionAccessIdentityResourceName = 'appDistributionAccessIdentity';
    sls.addDistributionAccessIdentity(distributionAccessIdentityResourceName, accessIdentity_1.generateDistributionAccessIdentity());
    const appBucketName = `${app.name}-app-bucket`;
    const appBucketResource = 'appBucket';
    sls.addS3Bucket(appBucketResource, bucket_1.generateS3Bucket(appBucketName));
    sls.addS3BucketPermission(appBucketResource + 'Permission', permission_1.generateS3BucketPermission(appBucketResource, distributionAccessIdentityResourceName));
    const distribution = new index_1.default({
        apiService: app.name,
        targets: [
            new s3Target_1.S3Target(appBucketName, distributionAccessIdentityResourceName, [
                { path: 'assets/*', ttl: 60 * 60 },
            ]),
            new gwTarget_1.GWTarget('apiGateway', [{ path: '*', ttl: 0 }]),
        ],
    }).json();
    sls.addDistribution('distribution', distribution);
    return sls.toYAML();
}
exports.default = generate;
//# sourceMappingURL=generate-yaml.js.map